#!/usr/bin/python
"""Hex tooling/utilities for operations."""
import math

_HEX = '0123456789abcdefABCDEF'
_HEX_LOOKUP = {x: int(x, 16) for x in (y + z for y in _HEX for z in _HEX)}

HEX_WHITE = "ffffff"
HEX_BLACK = "000000"
LIGHT_GREY = "lightgrey"
BLACK = "black"
DARK_GREY = "darkgrey"
WHITE = "WHITE"


def to_hex(rgb):
    """Convert a hex string 001122 -> tuple (0, 1, 2)."""
    return (_HEX_LOOKUP[rgb[0:2]],
            _HEX_LOOKUP[rgb[2:4]],
            _HEX_LOOKUP[rgb[4:6]])


def to_rgb_string(rgb):
    """Convert to rgb string."""
    return ('%02x%02x%02x' % rgb).lower()


def rgb_close(r1, g1, b1, r2, g2, b2):
    """Closest rgb check."""
    return math.sqrt(pow((r1 - r2) * 0.299, 2) +
                     pow((g1 - g2) * 0.587, 2) +
                     pow((b1 - b2) * 0.114, 2))


def is_rgb_string(value):
    """Sanity check against an rgb string."""
    if value is not None:
        lower = value.lower()
        check = [x for x in lower if x in ['a',
                                           'b',
                                           'c',
                                           'd',
                                           'e',
                                           'f',
                                           '0',
                                           '1',
                                           '2',
                                           '3',
                                           '4',
                                           '5',
                                           '6',
                                           '7',
                                           '8',
                                           '9']]
        return len(check) == 6
    return False



#!/usr/bin/python
"""
A formatter implementation for pygments.

Takes a text stream and converts to a cross stitch output (HTML).
"""

from pygments.formatter import Formatter
import pyxstitch.font as ft
import pyxstitch.formatter_layout as fl
import pyxstitch.hex as hu
import pyxstitch.symbols as sym
import pyxstitch.version as vers
from pyxstitch.output import PILFormat, TextFormat, MULTI_PAGE_OFF
from pyxstitch.config import Config
from pyxstitch.floss import Floss
from math import floor
from enum import Enum


class FormatterException(Exception):
    """Formatter exceptions."""


class CrossStitchFormatter(Formatter):
    """Formats output as a cross stitch pattern."""

    def __init__(self, **options):
        """Initialize the instance."""
        Formatter.__init__(self, **options)
        self._colors = {}
        self._default_color = hu.HEX_WHITE
        self.symbol_generator = sym.DefaultSymbolGenerator()
        self.font_factory = ft.Font().new_font_object()
        self.colorize = False
        self.dark = False
        self._lines = hu.LIGHT_GREY
        self._symbols = hu.BLACK
        self._light_symbol = hu.LIGHT_GREY
        self.file_name = None
        self.is_raw = False
        self._writer = None
        self.is_multipage = None
        self.config = None
        self.config_file = None
        self.is_bw = False
        self.vzoom = None
        self.hzoom = None
        self.floss = Floss()
        for token, style in self.style:
            if style['color']:
                self._colors[token] = style['color']

    def _get_backstitch_lines(self,
                              stitch,
                              x_start,
                              y_start,
                              x_end,
                              y_end,
                              offset):
        """Retrieve backstitch lines for a stitch."""
        x_st = x_start
        y_st = y_start
        x_en = x_end
        y_en = y_end
        newl = []
        if stitch in [ft.BackStitch.TopLeftBottomRight,
                      ft.BackStitch.TopLeft,
                      ft.BackStitch.BottomRight]:
            if stitch == ft.BackStitch.TopLeft:
                y_en = y_en - (offset / 2)
                x_en = x_en - (offset / 2)
            if stitch == ft.BackStitch.BottomRight:
                y_st = y_st + (offset / 2)
                x_st = x_st + (offset / 2)
            newl.append([x_st, y_st, x_en, y_en])
        if stitch in [ft.BackStitch.BottomLeftTopRight,
                      ft.BackStitch.BottomLeft,
                      ft.BackStitch.TopRight]:
            if stitch == ft.BackStitch.BottomLeft:
                y_st = y_st + (offset / 2)
                x_en = x_en - (offset / 2)
            if stitch == ft.BackStitch.TopRight:
                y_en = y_en - (offset / 2)
                x_st = x_st + (offset / 2)
            newl.append([x_st, y_en, x_en, y_st])
        if stitch in [ft.BackStitch.Left,
                      ft.BackStitch.TopLeftMid,
                      ft.BackStitch.BottomLeftMid,
                      ft.BackStitch.TopBottomMid]:
            x_en = x_start
            if stitch in [ft.BackStitch.TopLeftMid,
                          ft.BackStitch.TopBottomMid]:
                x_en = x_en + (offset / 2)
            if stitch in [ft.BackStitch.BottomLeftMid,
                          ft.BackStitch.TopBottomMid]:
                x_st = x_st + (offset / 2)
            newl.append([x_st, y_start, x_en, y_end])
        if stitch in [ft.BackStitch.Right,
                      ft.BackStitch.TopRightMid,
                      ft.BackStitch.BottomRightMid]:
            x_st = x_end
            if stitch == ft.BackStitch.TopRightMid:
                x_en = x_en - (offset / 2)
            if stitch == ft.BackStitch.BottomRightMid:
                x_st = x_st - (offset / 2)
            newl.append([x_st, y_start, x_en, y_end])
        if stitch == ft.BackStitch.Top:
            newl.append([x_start, y_start, x_end, y_start])
        if stitch == ft.BackStitch.Bottom:
            newl.append([x_start, y_end, x_end, y_end])
        return newl

    def map_color(self, input_color_map):
        """Map colors."""
        parts = input_color_map.lower().split("=")
        if len(parts) == 2:
            mapping = []
            idx = 0
            for p in parts:
                is_valid = hu.is_rgb_string(p)
                if not is_valid:
                    if idx == 0:
                        return False
                    else:
                        mapping.append(None)
                        break
                mapping.append(p)
                idx = idx + 1
            from_color = mapping[0]
            to_color = mapping[1]
            return self.floss.map(mapping[0], mapping[1])
        return False

    def _token_color(self, token):
        """We need to get the color for a token."""
        use_color = self._default_color
        if self.is_bw:
            use_color = hu.HEX_BLACK
        if token in self._colors:
            use_color = self._colors[token]
        use_hex = hu.to_hex(use_color)
        dmc = self.floss.lookup(use_hex)
        return fl.Style(dmc,
                        self.symbol_generator.next(dmc.name),
                        use_color)

    def _new_entry(self, ch, style):
        """new entry to process."""
        char = self.font_factory.get(ch)
        return (char, style, ch)

    def _legend(self, items, size):
        """Create legend chunk."""
        for i in range(0, len(items), size):
            yield items[i:i + size]

    def _init_output(self, outfile):
        """Init output."""

    def _zoom(self, pos, zoom):
        """Zoom check."""
        # we're zero-index before entering
        pos_one = pos + 1
        return pos_one < zoom[0] or pos_one > zoom[1]

    def format(self, tokensource, outfile):
        """Override to format."""
        if self.dark:
            self._symbols = hu.WHITE
            self._default_color = hu.HEX_BLACK
            self._light_symbol = hu.DARK_GREY
        if self.is_raw:
            self._writer = TextFormat()
        else:
            self._writer = PILFormat()
        entries = []
        current = []
        calc_height = 0
        calc_width = 0

        def _line(cur_height, cur_width, cur, entries):
            """Create a new line."""
            use_width = cur_width
            if len(cur) > use_width:
                use_width = len(cur)
            entries.append(cur)
            if len(cur) == 0:
                entries.append([self._new_entry(' ', styles)])
            return (cur_height + 1, use_width, [])

        for ttype, value in tokensource:
            while ttype not in self._colors:
                if ttype.parent is not None:
                    ttype = ttype.parent
                else:
                    break
            styles = self._token_color(ttype)
            if value is not None and '\n' in value and len(value) == 0:
                calc_height, calc_width, current = _line(calc_height,
                                                         calc_width,
                                                         current,
                                                         entries)
            else:
                for ch in value:
                    if ch == '\n':
                        calc_height, calc_width, current = _line(calc_height,
                                                                 calc_width,
                                                                 current,
                                                                 entries)
                    else:
                        current.append(self._new_entry(ch, styles))
        if len(current) > 0:
            calc_height += 1
            if len(current) > calc_width:
                calc_width = len(current)
            entries.append(current)
        calc_width += 1
        calc_width = ((calc_width * 2) +
                      (calc_width * max(self.font_factory.width())))
        mid_width = int(floor(calc_width / 2))
        calc_height += 1
        calc_height = (calc_height +
                       (calc_height * max(self.font_factory.height())))
        mid_height = int(floor(calc_height / 2))
        offset = 10
        cfg = Config(self.config, self.config_file)
        legend = 100
        legend_min = 500
        top_pad = 50
        left_pad = 50
        legend_req = legend_min + left_pad
        width_size = (calc_width * offset + left_pad)
        if width_size < legend_req:
            width_size = legend_req
        self._writer.init('RGB',
                          (width_size,
                           (calc_height * offset) + top_pad + legend),
                          hu.to_hex(self._default_color),
                          self.is_multipage,
                          cfg)
        self._writer.extras(cfg.dump())
        self._writer.text((offset, offset), vers.__version__, self._symbols)
        y = -1
        lines = []
        lgd = fl.Legend()
        for entry in entries:
            for height in self.font_factory.height():
                y += 1
                if self._zoom(y, self.vzoom):
                    continue
                x = -1
                grid = []
                has = False
                for cur, style, ch in entry:
                    dmc = style.dmc
                    color = self._symbols
                    first_write = True
                    if self.colorize:
                        color = dmc.name
                    for cell in cur.cells(height):
                        x += 1
                        if self._zoom(x, self.hzoom):
                            continue
                        if first_write:
                            self._writer.meta(cur.metadata(), style.save(), ch)
                            first_write = False
                        x_start = left_pad + offset + 0 + x * offset
                        y_start = top_pad + offset + 0 + y * offset
                        x_end = left_pad + offset + offset + x * offset
                        y_end = top_pad + offset + offset + y * offset
                        self._writer.rect([(x_start, y_start), (x_end, y_end)],
                                          outline=self._lines)
                        for stitch in cell:
                            lgd.add_raw_stitch(dmc)
                            if self.colorize:
                                color = '#' + dmc.rgb
                            lgd.add(style)
                            if isinstance(stitch, ft.BackStitch):
                                newl = self._get_backstitch_lines(stitch,
                                                                  x_start,
                                                                  y_start,
                                                                  x_end,
                                                                  y_end,
                                                                  offset)
                                for line in newl:
                                    lines.append(tuple(line + [color]))
                            if isinstance(stitch, ft.Stitch) or \
                               self.font_factory.is_backstitched:
                                if stitch == ft.Stitch.CrossStitch or \
                                   self.font_factory.is_backstitched:
                                    stitch_color = color
                                    if self.font_factory.is_backstitched:
                                        stitch_color = self._light_symbol
                                    lgd.add_raw_stitch(dmc)
                                    x_pos = left_pad + offset + 2 + x * offset
                                    self._writer.text((x_pos, y_start),
                                                      style.symbol,
                                                      stitch_color)
                        has = True
                if not has:
                    y -= 1
        # NOTE: we draw backstitch lines LAST to prevent overwrite
        for l in lines:
            self._writer.line((l[0], l[1], l[2], l[3]), fill=l[4])
        # add labels
        marked_widths = False
        for h in range(calc_height):
            if self._zoom(h, self.vzoom):
                continue
            if not marked_widths:
                marked_widths = True
                for w in range(calc_width):
                    if self._zoom(w, self.hzoom):
                        continue
                    if w % 10 == 0 or w == mid_width:
                        char = str(w)
                        if w == mid_width:
                            char = "X"
                        self._writer.text((left_pad + w * offset, top_pad - 5),
                                          char,
                                          self._symbols)
            if h % 10 == 0 or h == mid_height:
                char = str(h)
                if h == mid_height:
                    char = 'X'
                self._writer.text((left_pad - 5, top_pad + h * offset),
                                  char,
                                  self._symbols)
        chunk_idx = 0
        legend_tab = lgd.build()
        leg_height = (calc_height * offset) - (legend / 4)
        leg_height = leg_height + cfg.legend_hoff
        chunks = 8
        if self.is_multipage != MULTI_PAGE_OFF:
            chunks = 100
        for chunk in self._legend(legend_tab, chunks):
            leg_width = offset * 2 + (chunk_idx * legend_min)
            self._writer.legend((leg_width + cfg.legend_woff,
                                 leg_height),
                                "\n".join(chunk),
                                self._symbols)
            chunk_idx += 1
        self._writer.save(self.file_name)


def new_formatter(style,
                  file_name,
                  multipage,
                  colorize=False,
                  dark=False,
                  is_raw=False,
                  is_bw=False,
                  map_colors=None,
                  config=None,
                  font=None,
                  font_name=None,
                  rows=None,
                  columns=None,
                  symbols=None,
                  config_file=None,
                  horizontal=None,
                  vertical=None):
    """Create a new formatter."""
    formatting = CrossStitchFormatter(style=style)
    formatting.colorize = colorize
    formatting.dark = dark
    formatting.file_name = file_name
    formatting.is_multipage = multipage
    formatting.is_raw = is_raw
    formatting.is_bw = is_bw
    formatting.hzoom = horizontal
    formatting.vzoom = vertical
    if map_colors is not None and len(map_colors) > 0:
        for mapped in map_colors:
            if not formatting.map_color(mapped):
                raise FormatterException("unable to map: {}".format(mapped))
    if config is not None and len(config) > 0:
        formatting.config = config
    formatting.config_file = config_file
    if font_name is not None:
        formatting.font_factory = ft.Font().new_font_by_name(font_name,
                                                             rows=rows,
                                                             columns=columns)
    if symbols is not None:
        formatting.symbol_generator = sym.InputStringGenerator(symbols)
    return formatting



#!/usr/bin/python
"""Formatter layout settings."""


class Style(object):
    """Output/legend formatting."""

    def __init__(self, dmc, symbol, original_hex):
        """Init the instance."""
        self.dmc = dmc
        self.symbol = symbol
        self._raw_hex = original_hex

    def save(self):
        """Save to metadata format."""
        return [self.dmc.name, self.symbol, self._raw_hex]


class Legend(object):
    """Legend object."""

    def __init__(self):
        """Init the instance."""
        self._stitches = {}
        self._entries = []

    def add_raw_stitch(self, dmc):
        """Add raw stitch edge."""
        raw_color = dmc.name
        if raw_color not in self._stitches:
            self._stitches[raw_color] = 0
        self._stitches[raw_color] += 1

    def add(self, style):
        """New legend entry."""
        self._entries.append(style)

    def build(self):
        """Build output legend."""
        headers = []
        cols = ("dmc", "symbol", "edges", "floss")
        delim = []
        for col in cols:
            delim.append("---")
        headers.append(cols)
        headers.append(tuple(delim))
        output = []
        for item in self._entries:
            output.append(("{} ({})".format(item.dmc.name, item.dmc.rgb),
                           item.symbol,
                           self._stitches[item.dmc.name],
                           item.dmc.floss_number))
        return ["{:>40} {:>7} {:>6} {:>6}".format(x[0],
                                                  x[1],
                                                  x[2],
                                                  x[3])
                for x in headers + sorted(set(output))]



#!/usr/env/python
"""pyxstitch operating configuration settings."""
import os
import pyxstitch.log as log

_PAGE = "page_"
_NO_IDX = "no_index"
_LEGEND = "legend"
_FONT_SIZE = "font_size"
_ZEROS = [_PAGE + _NO_IDX, _PAGE + _LEGEND, _PAGE + _FONT_SIZE]
_LEGEND_ATTR = _LEGEND + "_"
_LGD_HOFF = _LEGEND_ATTR + "hoff"
_LGD_WOFF = _LEGEND_ATTR + "woff"
_OFFSET = [_LGD_HOFF, _LGD_WOFF]
_DELIMIT = "="


class Config(object):
    """Configuration definition."""

    def __init__(self, inputs, config_file):
        """Init the instance."""
        self.page_height = 600
        self.page_width = 1000
        self.page_pad = 50
        self.page_no_index = 0
        self.page_legend = 0
        self.legend_hoff = 0
        self.legend_woff = 0
        self.page_font_size = 0
        no_inputs = inputs is None or len(inputs) == 0
        if no_inputs:
            if config_file:
                self._parse_config(config_file)
        else:
            if config_file is not None:
                log.Log.write("config file ignored when given inputs")
            self._parse(inputs)

    def save(self):
        """Save to disk."""
        return [self.page_height,
                self.page_width,
                self.page_pad,
                self.page_no_index,
                self.page_legend,
                self.page_font_size]

    def dump(self):
        """dump extraneous settings passed in."""
        return [self.legend_hoff, self.legend_woff]

    @staticmethod
    def _create_page_input(key, value):
        """create an input."""
        return "{}{}{}{}".format(_PAGE, key, _DELIMIT, value)

    @staticmethod
    def load(values):
        """load config from saved type."""
        inputs = []
        if len(values) == 6:
            inputs.append(Config._create_page_input("height", values[0]))
            inputs.append(Config._create_page_input("width", values[1]))
            inputs.append(Config._create_page_input("pad", values[2]))
            inputs.append(Config._create_page_input(_NO_IDX, values[3]))
            inputs.append(Config._create_page_input(_LEGEND, values[4]))
            inputs.append(Config._create_page_input(_FONT_SIZE, values[5]))
        return Config(inputs, None)

    def _parse_config(self, conf):
        """Parse and load the config file."""
        if os.path.exists(conf):
            config_input = []
            with open(conf, 'r') as f:
                for l in f:
                    line = l.strip()
                    if line.startswith("#") or len(line) == 0:
                        continue
                    config_input.append(line)
            if len(config_input) > 0:
                self._parse(config_input)

    def _parse(self, inputs):
        """parse inputs."""
        for item in inputs:
            parts = item.split(_DELIMIT)
            if len(parts) != 2:
                log.Log.write('unable to parse config input: {}'.format(item))
                continue
            key = parts[0]
            val = parts[1]
            if key.startswith(_PAGE) or key.startswith(_LEGEND_ATTR):
                if key in dir(self):
                    try:
                        int_val = int(val)
                        if int_val > 0 or (int_val >= 0 and key in _ZEROS) or \
                           (int_val <= 0 and key in _OFFSET):
                            setattr(self, key, int_val)
                            continue
                    except Exception as e:
                        pass
            log.Log.write("invalid attribute {}".format(item))



#!/usr/bin/python
"""Defines a character for formatting text -> cross stitch character."""
from enum import IntEnum


class BadCharException(Exception):
    """Character is not defined or not fully defined."""


class Stitch(IntEnum):
    """Types of stitching."""

    CrossStitch = 1
    BottomLeft = 2
    BottomRight = 4
    TopLeft = 8
    TopRight = 16


class BackStitch(IntEnum):
    """Grid-based backstitching."""

    Top = 1
    Bottom = 2
    Left = 4
    Right = 8
    BottomLeftTopRight = 16
    TopLeftBottomRight = 32
    BottomLeft = 64
    BottomRight = 128
    TopLeft = 256
    TopRight = 512
    TopLeftMid = 1024
    TopRightMid = 2048
    BottomLeftMid = 4096
    BottomRightMid = 8192
    TopBottomMid = 16384


class Grid(object):
    """Defines a character grid of stitches."""

    def __init__(self):
        """Init a grid instance."""
        self.stitches = []


def _empty_grid(width, height):
    """Create an empty grid."""
    return [[Grid() for x in range(width)] for y in range(height)]


class Character(object):
    """Defines a cross stitch character."""

    def __init__(self, height, width):
        """Instance the character, creates an empty pattern grid."""
        self._height = height
        self._width = width
        self._pattern = _empty_grid(width, height)
        self.raw = None

    def cells(self, height):
        """Get stitches for  a specied character and returns them for it."""
        if height > self._height or height < 0:
            return None
        wide = self._pattern[height]
        for width in range(0, self._width):
            grid = wide[width]
            yield grid.stitches
        yield []

    def metadata(self):
        """Character metadata."""
        return [self._height, self._width]



#!/usr/bin/python
"""Simple/primitive logging."""


class Log(object):
    """Logging object."""

    is_verbose = True

    def writeln():
        """Write a blank line."""
        Log.write("")

    def write(message):
        """Write a logging line."""
        if Log.is_verbose:
            print(message)



#!/usr/env/python
"""Output formats."""
from collections import OrderedDict
from PIL import Image, ImageDraw
from io import StringIO
import pyxstitch.config as cfg
import pyxstitch.version as vers
import pyxstitch.log as log
import json
import os

RAW_FORMAT = "pyxstitch"

MULTI_PAGE_ON = "on"
MULTI_PAGE_OFF = "off"
MULTI_PAGE_AUTO = "auto"


class FormatError(Exception):
    """Format errors."""


class Format(object):
    """Base format output."""

    def init(self, style, dims, color, multipage, config):
        """init the instance."""
        raise FormatError("not implemented.")

    def rect(self, dims, outline=None):
        """Create  rectangle (square)."""
        raise FormatError("not implemented.")

    def text(self, dims, symbol, color):
        """Draw text."""
        raise FormatError("not implemented.")

    def line(self, dims, fill=None):
        """Draw a line."""
        raise FormatError("not implemented.")

    def save(self, file_name):
        """Save the output."""
        raise FormatError("not implemented.")

    def meta(self, char_meta, style, char):
        """Meta data."""
        raise FormatError("not implemented")

    def legend(self, pos, legend, symbols):
        """Legend output."""
        raise FormatError("not implemented")

    def extras(self, config_values):
        """Save extra information, not replayed."""
        raise FormatError("not implemented")


class PILFormat(Format):
    """PIL/image format."""

    _HTML = "<html><body>{}</body></html>"

    def __init__(self):
        """Init the output objects."""
        self._im = None
        self._dr = None
        self._img_dims = None
        self._is_multi = False
        self._img_color = None
        self._cfg = None
        self._legends = []
        self._legend_start = None

    def init(self, style, dims, color, multipage, config):
        """Init the image."""
        self._img_dims = dims
        self._img_color = color
        self._cfg = config
        if multipage == MULTI_PAGE_ON:
            self._is_multi = True
        elif multipage == MULTI_PAGE_AUTO:
            self._is_multi = self._img_dims[0] > self._cfg.page_width or \
                             self._img_dims[1] > self._cfg.page_height
        self._im = Image.new(style, dims, color)
        self._dr = ImageDraw.Draw(self._im)

    def rect(self, dims, outline=None):
        """Draw a rectangle."""
        self._dr.rectangle(dims, outline=outline)

    def text(self, dims, symbol, color):
        """Draw text."""
        self._dr.text(dims, symbol, color)

    def line(self, dims, fill=None):
        """Draw a line."""
        self._dr.line(dims, fill=fill)

    def _save(self, im, file_name):
        """Save an image."""
        log.Log.write("saving {}".format(file_name))
        im.save(file_name, quality=100)

    def extras(self, config_values):
        """Ignore extras on PIL."""
        pass

    def _calculate_width(self, offset):
        """Width calculation."""
        return self._cfg.page_width + offset * 2

    def _new_image(self, use_offset):
        """New output image."""
        return Image.new('RGB',
                         (self._calculate_width(use_offset),
                          self._cfg.page_height + use_offset * 2),
                         self._img_color)

    def _new_page(self, file_parts, page, report_out):
        """Create a new page name for file output."""
        paged = "{}_{}{}".format(file_parts[0],
                                 str(page).rjust(3, '0'),
                                 file_parts[1])
        report_out.append(os.path.basename(paged))
        return paged

    def save(self, file_name):
        """Save the output image."""
        if self._is_multi:
            page = 1
            width = self._img_dims[0]
            height = self._img_dims[1]
            use_height = min(height, self._cfg.page_height)
            use_width = min(width, self._cfg.page_width)
            use_offset = self._cfg.page_pad
            file_name_outputs = []

            file_parts = os.path.splitext(file_name)
            for h in range(0, height, use_height):
                if h > self._legend_start:
                    continue
                for w in range(0, width, use_width):
                    w_end = w + use_width
                    h_end = h + use_height
                    if w_end > width:
                        w_end = width
                    if h_end > height:
                        h_end = height
                    box = (w, h, w_end, h_end)
                    cropped = self._im.crop(box)
                    im = self._new_image(use_offset)
                    im.paste(cropped, (use_offset, use_offset))
                    extrema = im.convert("L").getextrema()
                    if extrema[0] == extrema[1]:
                        continue
                    paged = self._new_page(file_parts, page, file_name_outputs)
                    self._save(im, paged)
                    page += 1
            lgd_im = self._new_image(use_offset)
            lgd_dr = ImageDraw.Draw(lgd_im)
            use_lgd_pos = (use_offset, use_offset)
            self._legendize(lgd_im, lgd_dr, position=use_lgd_pos)
            self._save(lgd_im, self._new_page(file_parts,
                                              page,
                                              file_name_outputs))
            if self._cfg.page_no_index == 0:
                index_page = "{}-index.html".format(file_parts[0])
                log.Log.write("producing index page {}".format(index_page))
                with open(index_page, 'w') as f:
                    output_img = "".join(["<img src='{}' />".format(x)
                                         for x in file_name_outputs])
                    f.write(self._HTML.format(output_img))
        else:
            self._legendize(self._im, self._dr)
            self._save(self._im, file_name)

    def _legendize(self, image, draw, position=None):
        """Create a legend on a drawing."""
        for l in self._legends:
            if self._cfg.page_legend == 0:
                lgd = self._new_image(0)
                lgd_dr = ImageDraw.Draw(lgd)
                use_pos = position
                if use_pos is None:
                    pos = l[0]
                    use_pos = (int(pos[0]), int(pos[1]))
                lgd_dr.text((0, 0), l[1], l[2])
                width = self._calculate_width(0) + self._cfg.page_font_size
                percent = (width / float(lgd.size[0]))
                height = int((float(lgd.size[1]) * float(percent)))
                lgd = lgd.resize((width, height), Image.ANTIALIAS)
                image.paste(lgd, use_pos)
            else:
                log.Log.writeln()
                log.Log.write("legend")
                log.Log.write("======")
                log.Log.write(l[1])
                log.Log.writeln()

    def meta(self, char_meta, style, char):
        """Character metadata and style."""
        pass

    def legend(self, pos, legend, symbols):
        """Save legend information."""
        if self._legend_start is None or pos[1] < self._legend_start:
            self._legend_start = pos[1]
        self._legends.append((pos, legend, symbols))


class TextFormat(Format):
    """Raw texst output."""

    _TYPE = "type"
    _DATA = "data"
    _INIT = "init"
    _SAVE = "save"

    def __init__(self, dump=False):
        """Init the instance."""
        self._io = StringIO()
        self._dump = dump
        self._version = "0.5"

    def _unpack(self, args):
        """Unpack lists back to tuples (due to json)."""
        unpacked = []
        for arg in args:
            if isinstance(arg, list):
                unpacked.append(tuple(self._unpack(arg)))
            else:
                unpacked.append(arg)
        return unpacked

    def _log_replay(self, level, fmt, args):
        """log replay messages."""
        log.Log.write("{} -> {}".format(level, fmt.format(*args)))

    def replay(self, content, out_file_name):
        """Replay a file into another format."""
        line_idx = 1
        pil = PILFormat()
        for line in content.split("\n"):
            if len(line.strip()) == 0:
                continue
            try:
                obj = json.loads(line)
                cmd = obj[self._TYPE]
                datum = obj[self._DATA]
                if cmd == self._SAVE:
                    pil.save(out_file_name)
                    continue
                if cmd == self._INIT:
                    vers = datum[0]
                    if vers != self._version:
                        self._log_replay("warning",
                                         "Version: file {}, current {}",
                                         [vers, self._version])
                    conf = cfg.Config.load(datum[len(datum) - 1])
                    datum = datum[1:len(datum) - 1]
                    datum.append(conf)
                passing = self._unpack(datum)
                attr = getattr(pil, cmd)
                attr(*passing)
            except Exception as e:
                self._log_replay("error",
                                 "line number {}: {}",
                                 [line_idx, e])
                break
            line_idx += 1

    def init(self, style, dims, color, multipage, config):
        """Init the instance."""
        self._write(self._INIT, [self._version,
                                 style,
                                 dims,
                                 color,
                                 multipage,
                                 config.save()])

    def _write(self, obj_type, values):
        """Write data."""
        obj = {}
        obj[self._TYPE] = obj_type
        obj[self._DATA] = values
        obj = OrderedDict(sorted(obj.items(), reverse=True))
        self._io.write(json.dumps(obj) + "\n")

    def legend(self, pos, legend, symbols):
        """Legend output."""
        self._write("legend", [pos, legend, symbols])

    def rect(self, dims, outline=None):
        """Write rectangle information."""
        self._write("rect", [dims, outline])

    def text(self, dims, symbol, color):
        """Write text."""
        self._write("text", [dims, symbol, color])

    def line(self, dims, fill=None):
        """Write a line."""
        self._write('line', [dims, fill])

    def save(self, file_name):
        """Save outputs."""
        self._write(self._SAVE, [file_name])
        contents = self._io.getvalue()
        if self._dump:
            return contents
        with open(file_name, 'w') as f:
            f.write(contents)

    def extras(self, config_values):
        """Store extra information."""
        self._write("extras", [[vers.__version__, config_values]])

    def meta(self, char_meta, style, char):
        """Character metadata."""
        self._write("meta", [char_meta, style, char])



#!/usr/bin/python
"""Supports generating unique(ish) symbols for colors."""
import string


class SymbolError(Exception):
    """Symbol generation errors."""


class SymbolGenerator(object):
    """Generate symbols by color name."""

    def next(self, color):
        """Next color."""
        raise SymbolError("not implemented")


class DefaultSymbolGenerator(SymbolGenerator):
    """Default generator."""

    def __init__(self):
        """Create a new instance."""
        self._tracked = {}
        self._symbols = self._generate()
        self._idx = 0

    def _generate(self):
        """Generate the symbols."""
        syms = []
        syms = syms + sorted([x for x in string.printable if x.isalnum()])
        syms = syms + sorted([x for x in string.printable if not x.isalnum()])
        return syms

    def next(self, color):
        """Next color."""
        if color in self._tracked:
            return self._tracked[color]
        if self._idx >= len(self._symbols):
            raise SymbolError("out of symbols!")
        c = self._symbols[self._idx]
        self._idx += 1
        self._tracked[color] = c
        return c


class InputStringGenerator(DefaultSymbolGenerator):
    """Input string set of symbols."""

    def __init__(self, char_str):
        """Create a new instance."""
        if len(char_str) == 0:
            raise SymbolError("empty symbol input")
        if len(char_str) != len(set(char_str)):
            raise SymbolError("duplicate symbols detected")
        if len([x for x in char_str if not x.isalnum()]) > 0:
            raise SymbolError("non-alphanumeric not supported")
        self._characters = char_str
        super(InputStringGenerator, self).__init__()

    def _generate(self):
        """Generate symbols."""
        return self._characters



#!/usb/bin/python

"""Setup for pyxstitch."""

from setuptools import setup, find_packages
import os

__pkg_name__ = "pyxstitch"
with open(os.path.join(__pkg_name__, "version.py")) as v_file:
    exec(v_file.read())

long_description = ""
with open("README.rst", 'r') as f:
    long_description = f.read()

setup(
    author="Sean Enck",
    author_email="enckse@gmail.com",
    name=__pkg_name__,
    version=__version__,
    description='Convert source code to cross stitch patterns',
    long_description=long_description,
    url='https://github.com/enckse/pyxstitch',
    license='MIT',
    python_requires='>=3.5',
    packages=[__pkg_name__, __pkg_name__ + ".fonts"],
    install_requires=['pygments', 'Pillow'],
    keywords='crossstitch cross stitch',
    entry_points={
        'console_scripts': [
            'pyxstitch = pyxstitch.pyxstitch:main',
        ],
    },
)



#!/usr/bin/python
"""App for pyxstitch."""
from pygments import highlight
from pygments.lexers import get_lexer_for_filename, guess_lexer
from pygments.lexers import get_lexer_by_name
from pygments.styles import get_all_styles
import pyxstitch.formatter as fmt
import pyxstitch.output as out_fmt
import pyxstitch.font as fnt
import pyxstitch.version as vers
import pyxstitch.log as log
import argparse
import os
import sys
from pathlib import Path

_PNG = "png"
_RAW = out_fmt.RAW_FORMAT
_DEF_STYLE = "monokai"
_PDF = "pdf"
_TXT = ".txt"
_AUTODETECT = "autodetect"
_LIGHT = "light"
_DARK = "dark"
_DMC = "-dmc"
_LIGHT_DMC = _LIGHT + _DMC
_DARK_DMC = _DARK + _DMC
_B_AND_W = "bw"
_BANNER = 35
_BANNER_ONE = [17, 21, 25, 32]
_BANNER_TWO = [0, 1, 4, 6, 8, 10, 13, 14, 17, 25, 29, 30, 32]
_BANNER_THREE = [0, 2, 4, 6, 9, 13, 16, 17, 18, 21, 24, 25, 26, 28, 32, 33, 34]
_BANNER_FOUR = [0, 1, 5, 6, 8, 10, 12, 13, 17, 20, 21, 22, 25, 29, 30, 32, 34]
_BANNER_FIVE = [0, 6]
_BANNER_SIX = [0, 4, 5]


class InputArgs(object):
    """Input argumentss."""

    def __init__(self, args):
        """Init the input args."""
        for attr in dir(args):
            if attr.startswith("_"):
                continue
            setattr(self, attr, getattr(args, attr))
        self.is_raw = args.format == _RAW
        self.is_light = args.theme == _LIGHT
        self.default_style = args.style == _DEF_STYLE
        self.is_text = args.file == _TXT
        self.is_dark = args.theme.startswith(_DARK)
        self.is_dmc = args.theme.endswith(_DMC)
        self.invalid = False
        self._max = sys.maxsize
        self.hzoom = self._zoom("horizontal", args.hszoom, args.hezoom)
        self.vzoom = self._zoom("vertical", args.vszoom, args.vezoom)

    def _zoom(self, zoom_type, start, end):
        """Get zoom tuple for start/end."""
        u_start = 0
        u_end = self._max
        if start > 0:
            u_start = start
        if end > 0:
            u_end = end
        error_msg = None
        if start < 0 or end < 0:
            error_msg = "< 0"
        elif end < start:
            error_msg = "end < start"
        if error_msg:
            log.Log.write("invalid {} zoom request ({})".format(zoom_type,
                                                                error_msg))
            self.invalid = True
        return (u_start, u_end)


def _create_file_name(file_name, args):
    """Create output file names."""
    return "{}.{}".format(file_name, args.format)


def _print_spacer():
    """Print a spacer line."""
    _print_banner_line(range(0, _BANNER),
                       character='=',
                       leading="=",
                       trailing="=")


def _print_banner_line(marks, character='X', leading=" ", trailing=""):
    """Print a banner line."""
    actual = []
    for i in range(0, _BANNER):
        char = ' '
        if i in marks:
            char = character
        actual.append(char)
    log.Log.write(leading + "".join(actual) + trailing)


def _print_banner():
    """Print the pyxstitch banner."""
    _print_spacer()
    _print_banner_line(_BANNER_ONE)
    _print_banner_line(_BANNER_TWO)
    _print_banner_line(_BANNER_THREE)
    _print_banner_line(_BANNER_FOUR)
    _print_banner_line(_BANNER_FIVE)
    _print_banner_line(_BANNER_SIX)
    _print_spacer()


def _replay(args, file_name, content):
    """Do replay."""
    if args.is_raw:
        log.Log.write("can not replay raw to raw")
        exit(1)
    if not args.is_light or not args.default_style:
        log.Log.write('style and theme ignored during replay')
    out = args.output
    if args.output is None:
        out = _create_file_name(file_name, args)
    playback = out_fmt.TextFormat()
    playback.replay(content, out)
    exit(0)


def main():
    """Main-entry point."""
    default_font = fnt.Font().detect
    parser = argparse.ArgumentParser(
            description='Convert source code files to cross stitch patterns.')
    parser.add_argument('--file', help="""input source code file to create a
pattern from.""", type=str, default=_TXT)
    parser.add_argument('--lexer', help="""the source code lexer to use when
reading input and converting input tokens into colors""", type=str)
    parser.add_argument('--output', help="""output name""", type=str)
    parser.add_argument('--theme',
                        help="""pattern theme. indicates the background and
symbol color outputs for the output pattern""",
                        type=str,
                        default=_LIGHT,
                        choices=[_DARK,
                                 _LIGHT,
                                 _DARK_DMC,
                                 _LIGHT_DMC,
                                 _B_AND_W])
    parser.add_argument('--kv', metavar='N', help="""key/value configuration
parameters for setting the output format""", type=str, nargs='+')
    parser.add_argument('--map', help="""when selecting colors will map one
token color to another color (override)""", metavar='N', type=str, nargs='+')
    parser.add_argument('--config', help="""configuration file""", type=str)
    parser.add_argument('--multipage', help="""enable multipage output""",
                        type=str,
                        default=out_fmt.MULTI_PAGE_AUTO,
                        choices=[out_fmt.MULTI_PAGE_AUTO,
                                 out_fmt.MULTI_PAGE_ON,
                                 out_fmt.MULTI_PAGE_OFF])
    parser.add_argument('--format',
                        help="""output file format""",
                        type=str,
                        default=_PNG,
                        choices=[_PNG, _PDF, "jpg", _RAW])
    parser.add_argument('--style',
                        help="""token color styling (from pygments)""",
                        default=_DEF_STYLE,
                        choices=list(get_all_styles()))
    parser.add_argument('--font',
                        help="""font to use for the pattern""",
                        default=default_font,
                        choices=list(fnt.get_all_fonts()) + [default_font])
    parser.add_argument('--version',
                        action="store_true",
                        help="""display version""")
    parser.add_argument('--quiet',
                        action="store_true",
                        help="""less verbose output""")
    parser.add_argument('--symbols',
                        type=str,
                        help="""symbol set for stitching symbols""")
    parser.add_argument('--hszoom',
                        type=int,
                        default=0,
                        help="horizontal start zoom")
    parser.add_argument('--hezoom',
                        type=int,
                        default=0,
                        help="horizontal end zoom")
    parser.add_argument('--vszoom', type=int, default=0,
                        help="vertical start zoom")
    parser.add_argument('--vezoom', type=int, default=0,
                        help="vertical end zoom")
    args = parser.parse_args()
    if args.version:
        if not args.quiet:
            log.Log.writeln()
            _print_banner()
        log.Log.write(vers.__version__)
        if not args.quiet:
            log.Log.writeln()
        exit(0)
    if args.quiet:
        log.Log.is_verbose = False
    inputs = InputArgs(args)
    if inputs.invalid:
        exit(1)
    _run(inputs, default_font)


def _run(args, default_font):
    """Run pyxstitch."""
    content = None
    file_name = None
    file_ext = os.path.splitext(args.file)
    is_raw = False
    for item in file_ext:
        if item == "." + _RAW:
            is_raw = True
            break
    default_lexer = get_lexer_by_name("text")
    use_lexer = args.lexer
    use_style = args.style
    is_bw = False
    # Shortcut to black and white is to just use a Text lexer
    if args.theme == _B_AND_W:
        if use_lexer is not None:
            log.Log.write("black & white overrides lexer input")
        if use_style != _DEF_STYLE:
            log.Log.write("black & white overrides style")
        use_lexer = "Text"
        use_style = _B_AND_W
        is_bw = True
    is_auto = use_lexer == _AUTODETECT
    if is_auto:
        use_lexer = None
    if not is_raw:
        if use_lexer or args.is_text:
            if use_lexer:
                lexer = get_lexer_by_name(use_lexer)
            else:
                lexer = default_lexer
        else:
            try:
                lexer = get_lexer_for_filename(args.file)
            except Exception as e:
                log.Log.write(e)
                exit(1)
    if os.path.exists(args.file):
        file_name = file_ext[0]
        with open(args.file, 'r') as f:
            content = f.read()
    else:
        if file_ext[1] is not None and len(file_ext[1]) > 1:
            log.Log.write("file not found, pass extension for stdin or a file")
            exit(1)
        file_name = "output"
        content = "".join(sys.stdin.readlines())
    if is_raw:
        _replay(args, file_name, content)
    if is_auto:
        log.Log.write(content)
        try:
            lexer = guess_lexer(content)
            log.Log.write('using {} lexer'.format(lexer.name))
        except Exception as e:
            log.Log.write('unable to guess a lexer...defaulting to text')
            lexer = default_lexer
    output_name = args.output
    is_raw = args.is_raw
    if output_name is None:
        output_name = _create_file_name(file_name, args)
    else:
        if args.output.endswith(_RAW) and not is_raw:
            log.Log.write('specify output as {}?'.format(_RAW))
            exit(1)
    preproc = fnt.preprocess(content)
    text = preproc[0]
    rows = preproc[1]
    cols = preproc[2]
    config_file = args.config
    if config_file is None:
        home = str(Path.home())
        conf = os.path.join(home, ".pyxstitch.config")
        if os.path.exists(conf):
            config_file = conf
    if config_file is not None and not os.path.exists(config_file):
        log.Log.write("unable to find config file: {}".format(config_file))
    formatting = fmt.new_formatter(use_style,
                                   output_name,
                                   args.multipage,
                                   colorize=args.is_dmc,
                                   dark=args.is_dark,
                                   is_raw=is_raw,
                                   is_bw=is_bw,
                                   map_colors=args.map,
                                   font_name=args.font,
                                   rows=rows,
                                   columns=cols,
                                   symbols=args.symbols,
                                   config=args.kv,
                                   config_file=config_file,
                                   horizontal=args.hzoom,
                                   vertical=args.vzoom)
    if args.font == default_font:
        log.Log.write("font: {}".format(formatting.font_factory.display_name))
    log.Log.write("Using lexer: {}".format(lexer.name))
    highlight(text, lexer, formatting)


if __name__ == '__main__':
    main()



#!/usr/bin/python
"""Defines the default font functionality to get char -> stitch mappings."""
from pyxstitch.character import Character, BackStitch, Stitch, BadCharException


class FontException(Exception):
    """Font exceptions."""


class FontFactory(object):
    """Character stitching font creator."""

    def height(self):
        """Get the height of the font."""
        raise FontException("not implemented.")

    def get(self, character):
        """Get a character definition."""
        raise FontException("not implemented.")


def preprocess(content,
               replace={'\t': '    ', '\r': '\n', '\f': '\n', '\v': '\n'}):
    """Preprocess text/content before lexing."""
    val = content
    if replace is not None:
        for replacing in replace:
            val = val.replace(replacing, replace[replacing])
    parts = val.split("\n")
    cols = max([len(x) for x in parts])
    return (val, len(parts), cols)


class BaseFontFactory(FontFactory):
    """Base font factory."""

    def __init__(self):
        """Initialize the factory."""
        hw = self._height_width()
        self._height_offset = 2
        self._height = hw[0] + self._height_offset
        self._width = hw[1]
        self._top_off = 1
        self._bot_off = 1
        self._can_strip = False
        self.is_backstitched = False
        self._characters = self._initialize_characters()
        self.display_name = self._display()

    def _display(self):
        """Get the display name."""
        raise FontException("font does not declare a display name")

    def _monospace_ascii(self):
        """Generate a monospace-ascii name."""
        return self._create_name("monospace", "ascii")

    def _proportional_ascii(self):
        """Generate a proportional-ascii name."""
        return self._create_name("proportional", "ascii")

    def _create_name(self, spacing, charset):
        """Create a font name."""
        hw = self._height_width()
        return "{}-{}-{}x{}".format(spacing, charset, hw[1], hw[0])

    def _height_width(self):
        raise FontException("does not declare font height/width")

    def height(self):
        """get the font factory range of height for characters."""
        return range(self._height)

    def width(self):
        """get the font width."""
        return range(self._width)

    def get(self, ch):
        """Lookup a character in the font."""
        if ch in self._characters:
            return self._characters[ch]
        raise FontException("No font entry for character {}".format(ch))

    def _set_flags(self, val_str, enums, add_to):
        """Check which enum flags are set - append them to a list."""
        val = int(val_str)
        for e in enums:
            if e & val:
                for item in self._translate(e):
                    add_to.append(item)

    def _translate(self, enum_item):
        """Translate stitching types."""
        if isinstance(enum_item, Stitch):
            if enum_item == Stitch.BottomRight:
                return [BackStitch.BottomRight, BackStitch.BottomLeftTopRight]
            elif enum_item == Stitch.BottomLeft:
                return [BackStitch.BottomLeft, BackStitch.TopLeftBottomRight]
            elif enum_item == Stitch.TopLeft:
                return [BackStitch.TopLeft, BackStitch.BottomLeftTopRight]
            elif enum_item == Stitch.TopRight:
                return [BackStitch.TopRight, BackStitch.TopLeftBottomRight]
        return [enum_item]

    def _parse(self, definition, ch, can_strip):
        """Parse a character definition."""
        if definition is None:
            raise BadCharException("Invalid character definition")
        stripped = definition.strip()
        if len(stripped) == 0:
            raise BadCharException("Empty definition for character")
        parts = stripped.split("\n")
        has_height = len(parts)
        if has_height != self._height:
            if has_height != self._height - self._top_off - self._bot_off:
                raise BadCharException("Definition has an improper height")
            add_line = "|" + " | ".join(["" for x in range(self._width)]) + "|"
            for x in range(self._top_off):
                parts.insert(0, add_line)
            for x in range(self._bot_off):
                parts.append(add_line)
        height = 0
        if can_strip:
            empty_tracker = {}
        for part in parts:
            cur_height = height + self._height_offset
            defined = [x for x in part.split("|") if x != '']
            if len(defined) != self._width:
                raise BadCharException("Definition has an improper width")
            width = 0
            for entry in defined:
                if len(entry.strip()) > 0:
                    section = entry.split(".")
                    adding = []
                    self._set_flags(section[0], Stitch, adding)
                    self._set_flags(section[1], BackStitch, adding)
                    ch[height][width].stitches = adding
                else:
                    if can_strip and height > 0 and cur_height <= self._height:
                        if width not in empty_tracker:
                            empty_tracker[width] = []
                        empty_tracker[width].append(width)
                width += 1
            height += 1
        new_width = self._width
        if can_strip:
            for w in reversed(range(0, self._width)):
                if w in empty_tracker:
                    heights = len(empty_tracker[w])
                    if heights == self._height - self._height_offset:
                        new_width -= 1
                    else:
                        break
                else:
                    break
        return (ch, new_width)

    def _drop_lines(self, write_to, ch, chars, lines):
        """Drop lines from an existing character."""
        raw = chars[ch].raw
        raw_lines = raw.split("\n")
        idx = 0
        conv = []
        for line in raw_lines:
            idx = idx + 1
            if idx in lines:
                continue
            conv.append(line)
        write_to[ch] = self._build_character("\n".join(conv))

    def _build_character(self, stitching, char_strip=True):
        """Build a character into an object definition."""
        ch = Character(self._height, self._width)
        try:
            do_strip = self._can_strip and char_strip
            pattern = self._parse(stitching,
                                  ch._pattern,
                                  do_strip)
            if do_strip:
                ch = Character(self._height, pattern[1])
            ch._pattern = pattern[0]
            ch.raw = stitching
        except BadCharException as e:
            raise BadCharException("{} -> {}".format(str(e), stitching))
        return ch

    def _initialize_characters(self):
        """Initialize char set."""
        raise FontException("characeter set not defined")


def get_all_fonts():
    """Get all fonts."""
    return Font().get_names()


class Font(object):
    """Create font instance."""

    detect = "detect"

    def __init__(self):
        """Init the font creation object."""
        from pyxstitch.fonts.five_by_nine_mono import FiveByNineMono
        from pyxstitch.fonts.three_by_five_mono import ThreeByFiveMono
        from pyxstitch.fonts.three_by_seven_mono import ThreeBySevenMono
        from pyxstitch.fonts.two_by_five_mono import TwoByFiveMono
        from pyxstitch.fonts.three_by_six_prop import ThreeBySixProp
        from pyxstitch.fonts.four_by_seven_prop import FourBySevenProp
        from pyxstitch.fonts.five_by_nine_prop import FiveByNineProp
        self._supported_types = [FiveByNineMono,
                                 ThreeBySevenMono,
                                 TwoByFiveMono,
                                 ThreeByFiveMono,
                                 ThreeBySixProp,
                                 FourBySevenProp,
                                 FiveByNineProp]
        self._names = {}
        self._instance_cache = []
        # tuple is (index, column threshold, row threshold)
        self._add_instance(FiveByNineMono, (0, 31, 21))
        self._add_instance(ThreeBySevenMono, (1, 46, 26))
        self._add_instance(TwoByFiveMono, (2, None, None))
        self._add_instance(ThreeByFiveMono, (3, 46, 31))
        self._add_instance(ThreeBySixProp, (4, None))
        self._add_instance(FourBySevenProp, (5, None))
        self._add_instance(FiveByNineProp, (6, None))

    def _add_instance(self, use_type, auto_settings):
        """Add a cached instance."""
        inst = use_type()
        self._names[inst.display_name] = auto_settings
        self._instance_cache.append(inst)

    def get_names(self):
        """Get font names."""
        return self._names.keys()

    def new_font_by_name(self, font_name, rows=None, columns=None):
        """Get a font by name."""
        use_name = font_name
        if font_name == self.detect:
            dflt = None
            if rows is not None and columns is not None:
                for detect in sorted(self._names.keys()):
                    vals = self._names[detect]
                    if len(vals) > 2:
                        c = vals[1]
                        r = vals[2]
                        if c is None and r is None:
                            dflt = detect
                            continue
                        if rows < r and columns < c:
                            use_name = detect
                if use_name == font_name:
                    use_name = dflt
            else:
                raise FontException("requires dimensions (rows, cols)")
        if use_name in self._names:
            typed = self._supported_types[self._names[use_name][0]]
            inst = self.new_font_object(typed)
            return inst
        else:
            raise FontException("unknown font name: {}".format(font_name))

    def new_font_object(self, font_type=None, columns=None):
        """Create a new font object."""
        use_type = None
        if font_type is None:
            use_type = self._supported_types[0]
        if font_type in self._supported_types:
            use_type = font_type
        in_error = use_type is None
        if not in_error:
            obj = [x for x in self._instance_cache if use_type == x.__class__]
            if len(obj) == 1:
                return obj[0]
            else:
                in_error = True
        if in_error:
            raise FontException("unknown font type: {}".format(font_type))
